# popular ui submodule

import pygame
import vector
import face
import thoughts
from colors import *
from popular import Gossip

# config
log = True

SKIP_TITLE = False

draw_affinity = False
# draw aggregate popularity (bind: F2)
draw_total_pop = False
# draw reaction previews while building gossip (bind: F3)
draw_gossip_result_preview = False


DESK_WIDTH = 160
DESK_DEPTH = 48
DESK_HEIGHT = 48
# space between desks
DESK_DISTANCE_X = 64
DESK_DISTANCE_Y = 80

BUTTON_WIDTH = 128
BUTTON_HEIGHT = 32

BOX_WIDTH = 96
BOX_HEIGHT = 96
BOX_DISTANCE = 64

#
# strings
#

TITLE_TEXT = 'POPULAR'
TITLE_GOAL_TEXT = 'Become the most popular kid in class.'
GOSSIP_HELP = 'Drag students into the boxes to create some gossip.'
SEND_HELP = 'Click students to define the route your note will take.'

# transient variables
# the importing module should set these!
mouse_x, mouse_y = 0, 0
time = 0
ui = None
debug_font = None
debug_font_small = None
students = []
player = None
screen = None

#
# helper functions
#

def grid_to_screen(grid_x, grid_y, offset=(0, 0)):
    """
    returns top left of desk, in screen pixels, from specified point
    on "desk grid", plus optional offset
    """
    x = grid_x * (DESK_DISTANCE_X + DESK_WIDTH)
    x += DESK_DISTANCE_X
    x += offset[0]
    y = grid_y * (DESK_DISTANCE_Y + DESK_HEIGHT + DESK_DEPTH)
    y += DESK_DISTANCE_Y
    y += offset[1]
    return (x, y)


def affinity_color(aff, aff_base=10, color_base=64):
    """
    returns a color scaled with given affinity value
    
    aff_base = baseline (because affinity is unbounded)
    color_base = starting brightness of result color
    """
    v = 255 * (aff / float(aff_base))
    v = abs(int(v)) + 128
    v = min(255, v)
    c = pygame.Color(color_base)
    if aff < 0:
        c.r = v
    elif aff > 0:
        c.g = v
    return c


#
# UI classes
#

# ui states
UI_TITLE = 0
# starting, look-only state
UI_VIEW = 1
# constructing gossip
UI_GOSSIP = 2
# building a note pass route
UI_NOTEROUTE = 3
# note is moving
UI_PASSING = 4
# student reacts to a note
UI_PASS_REACTING = 5


class ClassroomUI():
    tooltip_color = BLACK
    tooltip_bgcolor = LIGHT_GREY
    # how much bigger the tooltip box should be than its text
    tooltip_border = 4
    
    def __init__(self):
        self.status_line = UIStatusLine(self)
        self.cgb = CreateGossipButton(self)
        self.sgb = SendGossipButton(self)
        self.pgb = PassGossipButton(self)
        self.box_spk = GossipSpeakerBox(self)
        self.box_sub = GossipSubjectBox(self)
        self.box_like = GossipLikeHateBox(self)
        self.box_obj = GossipObjectBox(self)
        self.rank_bar = AffinityRankingBar(self)
        self.elements = [self.status_line, self.cgb, self.sgb, self.pgb, self.box_spk, self.box_sub, self.box_like, self.box_obj, self.rank_bar]
        # state-related variables
        self.drag_elements = []
        self.hover_elements = []
        self.can_input = True
        self.tooltip_text = None
        self.gossip = None
        self.note = None
        self.selected_student = None
        self.focus = vector.Vector2D(0, 0)
        self.dialog = None
        if SKIP_TITLE:
            self.state_view()
        else:
            self.state_title()
    
    def state_title(self):
        self.state = UI_TITLE
    
    def state_view(self):
        "restarts turn loop after a gossip->pass->react cycle completes"
        self.state = UI_VIEW
        self.status_line.text = None
        # blank data objects from previous turn
        self.gossip = None
        self.note = None
        # restore button states
        self.cgb.active = True
        self.pgb.active = False
        self.box_spk.active = False
        self.box_sub.active = False
        self.box_sub.contents = None
        self.box_like.active = False
        self.box_like.likes = True
        # update "like box" text as well so data and viz don't scramble
        self.box_like.update_text()
        self.box_obj.active = False
        self.box_obj.contents = None
        self.can_input = True
        # clear all student /face/ reactions so they won't draw next turn
        for s in students:
            s.face.reaction = None
    
    def state_gossip(self, instigator):
        # begin constructing an item of gossip
        self.state = UI_GOSSIP
        self.status_line.text = GOSSIP_HELP
        self.cgb.active = False
        # send box only becomes active when boxes contain valid gossip
        self.box_spk.active = True
        self.box_sub.active = True
        self.box_like.active = True
        self.box_obj.active = True
        # contents of speaker box should be actual speaker
        self.box_spk.contents = instigator.face
    
    def state_send(self):
        # send button is only active when valid gossip exists,
        # which in turn is checked in update()
        self.state = UI_NOTEROUTE
        self.status_line.text = SEND_HELP
        self.sgb.active = False
        # don't enable "pass" button until route has at least one point
        #self.pgb.active = True
        self.note = GossipNote(self.gossip)
        if log: "player started building a route."
    
    def state_passing(self, first=False):
        self.state = UI_PASSING
        self.status_line.text = None
        self.status_line.speaker = None
        self.pgb.active = False
        self.note.set_next_dest_in_route()
        if log: "note began passing."
        self.can_input = False
    
    def state_reacting(self):
        # not much happens here, GossipNote knows and does
        # most of the relevant stuff for this state
        self.state = UI_PASS_REACTING
        self.can_input = True
        reacting_student = self.note.route.pop(0)
        # if we are on the speaker (ie where the note starts) skip over them
        if reacting_student is self.note.gossip.speaker:
            self.state_passing()
        else:
            # reaction used to be on status line, now in a dialog
            reacting_student.face.start_reacting()
            self.dialog = ReactionDialog(reacting_student.reaction)
    
    def finished_reacting(self):
        self.dialog = None
        # callback from StudentFace OR from input() if no reaction time delay
        if len(self.note.route) == 0:
            self.state_view()
        else:
            self.state_passing()
    
    def hovering_class(self, hover_class_list):
        "returns first instance of the specified class we're hovering"
        # accept only a list of classes
        #is_list = hover_class.__name__ is list
        for e in self.hover_elements:
            if e.__class__ in hover_class_list:
                return e
        return None
    
    def update_elements(self):
        # TODO: consider the benefits of checking hover/unhover
        # in the update() of each element
        # (nothing much happens in elements' update() currently)
        
        # clear hover elements, remember from last frame
        was_hovering = self.hover_elements[:]
        self.hover_elements = []
        # save lists of elements that we just started or finished hovering,
        # as some of them will on a complete self.hover_elements for their
        # hover() and unhover() to work properly
        just_hovered = []
        just_unhovered = []
        # run events on relevant elements
        for e in self.elements:
            if not e.active:
                continue
            # see what we're hovering over
            if e.bounds.top < mouse_y < e.bounds.bottom and \
                e.bounds.left < mouse_x < e.bounds.right:
                    self.hover_elements.append(e)
                    # did we just start hovering this frame?
                    if not e in was_hovering:
                        just_hovered.append(e)
            # did we just stop hovering this frame?
            elif e in was_hovering:
                just_unhovered.append(e)
        # process the elements whose status has changed
        for e in just_hovered:
            e.hover()
        for e in just_unhovered:
            e.unhover()
        # every active element gets to update
        # (safer to call this at the very end after hover() etc)
        for e in self.elements:
            if e.active:
                e.update()
    
    def update_selected_student(self):
        """
        asks the UI to determine who the selected student (if any) might
        be - usually called from the unhover() of a given element
        """
        self.selected_student = None
        # if a note is passing, students follow it because we set
        # mousex/y in input(), but don't count that as player hovering
        if self.state in [UI_PASSING, UI_PASS_REACTING]:
            return
        # if bounds of any 2 elements overlap, we might still be hovering
        # another student, set them as selected if so
        # handle direct hover or GossipBox differently
        sf = self.hovering_class([StudentFace])
        if sf:
            self.selected_student = sf.student
            return
        sf = self.hovering_class(gossip_box_classes)
        if sf and sf.contents:
            self.selected_student = sf.contents.student
    
    def update_gossip(self):
        if not self.box_sub.contents or not self.box_obj.contents:
            self.gossip = None
            return
        spk = player
        sub = self.box_sub.contents.student
        likes = self.box_like.likes
        obj = self.box_obj.contents.student
        # only create new gossip item if ours is out of date
        if self.gossip and self.gossip.likes == likes and \
            self.gossip.speaker is spk and self.gossip.subject is sub and \
            self.gossip.object is obj:
            return
        self.gossip = Gossip(spk, sub, likes, obj)
    
    def update_focus(self):
        # in most cases, focus on mouse cursor
        self.focus.x, self.focus.y = mouse_x, mouse_y
        # if selecting a student (via direct hover or GossipBox), focus on them
        if not self.selected_student:
            return
        box = self.hovering_class(gossip_box_classes)
        if box:
            self.focus = vector.Vector2D(box.contents.bounds.center)
    
    def update(self):
        # hide cursor if we're not accepting input
        pygame.mouse.set_visible(self.can_input)
        if self.note:
            self.note.update()
        self.update_selected_student()
        self.update_elements()
        self.update_gossip()
        self.update_focus()
        # only show "send" button if gossip is fully formed
        if self.state == UI_GOSSIP and self.gossip:
            self.sgb.active = True
        else:
            self.sgb.active = False
        # only show "pass" button if note's route is valid
        if self.state == UI_NOTEROUTE:
            if len(self.note.route) > 1:
                self.pgb.active = True
            else:
                self.pgb.active = False
        # do we have a tooltip to draw?
        self.tooltip_text = None
        for e in self.hover_elements:
            if e.tooltip_text:
                self.tooltip_text = e.tooltip_text
    
    def click(self):
        if self.state == UI_TITLE:
            self.state_view()
            return
        # special case: when reacting, click to continue passing
        if self.state == UI_PASS_REACTING:
            self.finished_reacting()
            return
        for e in self.hover_elements:
            e.click()
    
    def unclick(self):
        for e in self.hover_elements:
            e.unclick()
        # drop all dragged items(?)
        self.drag_elements = []
    
    def draw_tooltip(self):
        x = mouse_x + debug_font.get_height() / 2
        y = mouse_y - debug_font.get_height() / 2
        text_surf = debug_font.render(self.tooltip_text, True, self.tooltip_color)
        w = text_surf.get_width() + self.tooltip_border
        h = text_surf.get_height() + self.tooltip_border
        text_rect = pygame.Rect(x, y, w, h)
        pygame.draw.rect(screen, self.tooltip_bgcolor, text_rect)
        x += self.tooltip_border / 2
        y += self.tooltip_border / 2
        screen.blit(text_surf, (x, y))
    
    def draw_title(self):
        screen.fill(BLACK)
        x = 64
        y = screen.get_height() / 2
        title_surf = debug_font.render(TITLE_TEXT, True, LIGHT_GREY)
        screen.blit(title_surf, (x, y))
        y += title_surf.get_height() * 2
        goal_surf = debug_font.render(TITLE_GOAL_TEXT, True, LIGHT_GREY)
        screen.blit(goal_surf, (x, y))
    
    def draw(self):
        if self.state == UI_TITLE:
            self.draw_title()
            return
        for e in self.elements:
            if e.active:
               e.draw()
        if self.note:
            self.note.draw()
        if self.dialog:
            self.dialog.draw()
        # draw tooltips last
        if self.tooltip_text:
            self.draw_tooltip()


class UIElement():
    active = False
    width, height = 64, 64
    x, y = 0, 0
    tooltip_text = None
    
    def __init__(self, ui):
        self.ui = ui
        self.active = True
        self.r = pygame.Rect(self.x, self.y, self.width, self.height)
        self.bounds = self.r
    
    def update(self):
        pass
    
    def hover(self):
        pass
    
    def unhover(self):
        pass
    
    def click(self):
        pass
    
    def unclick(self):
        pass
    
    def draw(self):
        pass


class UIButton(UIElement):
    width, height = BUTTON_WIDTH, BUTTON_HEIGHT
    text = 'DEBUG'
    text_color = BLACK
    tooltip_color = BLACK
    default_color = MID_GREY
    hover_color = LIGHT_GREY
    click_color = WHITE
    filled = True
    caption_x = BUTTON_WIDTH / 2
    caption_y = BUTTON_HEIGHT / 2
    
    def __init__(self, ui):
        UIElement.__init__(self, ui)
        self.color = self.default_color
    
    def hover(self):
        self.color = self.hover_color
    
    def unhover(self):
        self.color = self.default_color
    
    def click(self):
        self.color = self.click_color
    
    def unclick(self):
        self.color = self.hover_color
    
    def draw(self):
        if self.filled:
            pygame.draw.rect(screen, self.color, self.r)
        else:
            pygame.draw.rect(screen, self.color, self.r, 2)
        text_surf = debug_font.render(self.text, True, self.text_color)
        # draw caption
        x = self.caption_x + self.x
        y = self.caption_y + self.y
        x -= text_surf.get_width() / 2
        y -= text_surf.get_height() / 2
        screen.blit(text_surf, (x, y))


class CreateGossipButton(UIButton):
    x = 16
    text = 'GOSSIP'
    tooltip_text = 'Start a new piece of Gossip'
    
    def __init__(self, ui):
        self.y = screen.get_height() - BUTTON_HEIGHT - 16
        UIButton.__init__(self, ui)
    
    def unclick(self):
        UIButton.unclick(self)
        # if this button is pressed, we know it's player who is gossiping
        self.ui.state_gossip(player)


class SendGossipButton(UIButton):
    text = 'SEND'
    
    def __init__(self, ui):
        self.x = screen.get_width() - BUTTON_WIDTH - 16
        self.y = screen.get_height() - BUTTON_HEIGHT - 16
        UIButton.__init__(self, ui)
    
    def unclick(self):
        UIButton.unclick(self)
        self.ui.state_send()


class PassGossipButton(UIButton):
    text = 'PASS'
    
    def __init__(self, ui):
        self.x = screen.get_width() - BUTTON_WIDTH - 16
        self.y = screen.get_height() - BUTTON_HEIGHT - 16
        UIButton.__init__(self, ui)
    
    def unclick(self):
        UIButton.unclick(self)
        # tell ui this is the first pass of a note (slight special case)
        self.ui.state_passing(True)


class GossipBox(UIButton):
    width, height = BOX_HEIGHT, BOX_WIDTH
    text = 'GOSSIP BOX'
    text_color = LIGHT_GREY
    caption_x = BOX_WIDTH / 2
    filled = False
    contents = None
    
    def __init__(self, ui):
        self.y = screen.get_height() - (BOX_DISTANCE * 2)
        self.caption_y = -debug_font.get_height() / 2
        UIButton.__init__(self, ui)
    
    def update(self):
        # change our tooltip depending on our state
        if self.contents:
            self.tooltip_text = self.filled_tooltip_text
        else:
            self.tooltip_text = self.empty_tooltip_text
    
    def hover(self):
        UIButton.hover(self)
        # TODO: show everyone's reaction to this student for current gossip!
    
    def unhover(self):
        UIButton.unhover(self)
        self.ui.update_selected_student()
    
    def unclick(self):
        for e in ui.drag_elements:
            if e.__class__ is StudentFace:
                self.contents = e
                # ui will remove e from drag_elements
    
    def draw(self):
        UIButton.draw(self)
        if not self.contents:
            return
        # draw student face if we have one
        x = self.x + self.width / 2
        y = self.y + self.height / 2.5
        self.contents.draw_face_thumbnail((x, y), True, (0.5, 0.5), True, WHITE)


class GossipSpeakerBox(GossipBox):
    x = BOX_DISTANCE
    text = 'Speaker'
    empty_tooltip_text = 'Student spreading this Gossip'
    filled_tooltip_text = empty_tooltip_text
    
    def unclick(self):
        # can't drag stuff here unlike subj/obj
        pass
    
    def draw(self):
        GossipBox.draw(self)
        # "says that" text to right of box
        text_surf = debug_font.render("SAYS:", True, self.text_color)
        x = self.x + BOX_WIDTH + BOX_DISTANCE
        x -= text_surf.get_width() / 2
        y = self.y + BOX_HEIGHT / 2
        y -= text_surf.get_height() / 2
        screen.blit(text_surf, (x, y))


class GossipSubjectBox(GossipBox):
    x = BOX_WIDTH + BOX_DISTANCE * 3
    text = 'Subject'
    empty_tooltip_text = 'Drag a Student here'
    filled_tooltip_text = ''


class GossipLikeHateBox(UIButton):

    text = 'LIKES'
    tooltip_text = 'Toggle Like or Hate'
    text_color = BLACK
    filled = True
    likes = True
    
    def __init__(self, ui):
        self.x = BOX_WIDTH * 2 + BOX_DISTANCE * 4
        self.y = screen.get_height() - (BOX_DISTANCE * 2)
        self.width, self.height = GossipBox.width, GossipBox.height
        self.caption_x = GossipBox.caption_x
        self.caption_y = BOX_HEIGHT / 2
        UIButton.__init__(self, ui)
    
    def update_text(self):
        if self.likes:
            self.text = 'LIKES'
        else:
            self.text = 'HATES'
    
    def unclick(self):
        UIButton.unclick(self)
        self.likes = not self.likes
        self.update_text()


class GossipObjectBox(GossipBox):
    x = BOX_WIDTH * 3 + BOX_DISTANCE * 5
    text = 'Object'
    empty_tooltip_text = 'Drag a Student here'
    filled_tooltip_text = ''


gossip_box_classes = [GossipSpeakerBox, GossipSubjectBox, GossipObjectBox]


class UIStatusLine(UIElement):
    "status line displaying help for current mode, or student thoughts"
    x, y = 0, GossipBox.y - BOX_HEIGHT * 0.8
    text_color = WHITE
    
    def __init__(self, ui):
        UIElement.__init__(self, ui)
        self.text = ''
        self.speaker = None
        self.width, self.height = screen.get_width(), BUTTON_HEIGHT
    
    def draw(self):
        # draw rectangle
        pygame.draw.rect(screen, DARK_GREY, self.r)
        if not self.text:
            return
        # draw text
        # make room for mini student face if reacting
        if ui.state == UI_PASS_REACTING:
            text_offset_x = student_head_rect.width / 2
            self.speaker.face.draw_face_thumbnail((self.x, self.y), False)
        else:
            text_offset_x = 8
        text_surf = debug_font.render(self.text, True, self.text_color)
        y = self.y + text_surf.get_height() / 3
        screen.blit(text_surf, (self.x + text_offset_x, y))


class ReactionDialog(UIElement):
    
    "window jam-packed full of info on a student's reaction to gossip"
    
    width, height = 650, 650
    bg_color = LIGHT_GREY
    
    def __init__(self, reaction):
        self.reaction = reaction
        self.gossip = reaction.gossip
        self.listener = reaction.student
        self.x = (screen.get_width() - self.width) / 2
        self.y = (screen.get_height() - self.height) / 2
        self.bounds = pygame.Rect(self.x, self.y, self.width, self.height)
        # get relevant strings now;
        # nothing more expensive than arith or property access in draw()
        self.reaction_thoughts = self.reaction.thoughts()
        self.truth = self.reaction.truthiness
        speaker_aff = self.listener.affinity[self.gossip.speaker]
        self.speaker_thoughts = thoughts.speaker_affinities[speaker_aff]
        self.speaker_thoughts = thoughts.parse(self.speaker_thoughts, self.gossip)
        # surface we can blit to dim the screen behind us
        bg_surf = pygame.Surface((1, 1))
        bg_surf.set_alpha(128)
        self.bg_dim_surf = pygame.transform.scale(bg_surf, screen.get_size())
    
    def draw_background(self):
        # dim screen behind dialog
        screen.blit(self.bg_dim_surf, (0, 0))
        # background
        pygame.draw.rect(screen, self.bg_color, self.bounds)
    
    def draw_titlebar(self):
        x = self.x + 40
        y = self.y + 40
        # listener face and name
        l_text = "%s hears:" % self.listener
        l_txt_surf = debug_font.render(l_text, True, BLACK)
        self.listener.face.draw_face_thumbnail((x, y), True, (0.75, 0.75))
        x += 60
        screen.blit(l_txt_surf, (x, y))
    
    def draw_gossip(self):
        spk = self.gossip.speaker
        subj = self.gossip.subject
        obj = self.gossip.object
        # gossip contents
        x = self.x + 130
        y = self.y + 60
        bub_height = reaction_bubble_img.get_height() / 2
        y += bub_height
        # speaker
        spk.face.draw_face_thumbnail((x, y), True, (0.5, 0.5), True)
        # speaker affinity
        spk_aff = self.listener.affinity[spk]
        spk_aff_string = thoughts.affinity_descriptions[spk_aff]
        if spk is self.listener:
            spk_aff_string = 'Me'
        spk_aff_color = affinity_color(spk_aff, 15)
        spk_aff_surf = debug_font_small.render(spk_aff_string, True, spk_aff_color)
        x_off = spk_aff_surf.get_width() / 2
        y_off = 45
        screen.blit(spk_aff_surf, (x - x_off, y + y_off))
        
        x += 40
        screen.blit(reaction_bubble_img, (x, y - bub_height))
        # subject
        x += 90
        subj.face.draw_face_thumbnail((x, y), True, (0.5, 0.5), True)
        # subject affinity
        sub_aff = self.listener.affinity[subj]
        sub_aff_string = thoughts.affinity_descriptions[sub_aff]
        if subj is self.listener:
            sub_aff_string = 'Me'
        sub_aff_color = affinity_color(sub_aff, 15)
        sub_aff_surf = debug_font_small.render(sub_aff_string, True, sub_aff_color)
        x_off = sub_aff_surf.get_width() / 2
        screen.blit(sub_aff_surf, (x - x_off, y + y_off))
        
        x += 90
        # likes/hates
        screen.blit([hate_img, love_img][self.gossip.likes], (x, y - 32))
        x += 110
        # object
        self.gossip.object.face.draw_face_thumbnail((x, y), True, (0.5, 0.5), True)
        # object affinity
        obj_aff = self.listener.affinity[obj]
        obj_aff_string = thoughts.affinity_descriptions[obj_aff]
        if obj is self.listener:
            obj_aff_string = 'Me'
        obj_aff_color = affinity_color(obj_aff, 15)
        obj_aff_surf = debug_font_small.render(obj_aff_string, True, obj_aff_color)
        x_off = obj_aff_surf.get_width() / 2
        screen.blit(obj_aff_surf, (x - x_off, y + y_off))
    
    def draw_thoughts_bg(self):
        x = self.x + 40
        y = self.y + 260
        screen.blit(thotbub_huge_img, (x - 60, y - 40))
        self.listener.face.draw_face_thumbnail((x, y), True, (0.75, 0.75))
        x += 60
        thinker_str = '%s thinks:' % self.listener
        thinker_txt_surf = debug_font.render(thinker_str, True, BLACK)
        screen.blit(thinker_txt_surf, (x, y))
    
    def draw_gut_reaction(self):
        # gut reaction
        x = self.x + 40
        y = self.y + 320
        reac_label_surf = debug_font_small.render('Gut reaction:', True, BLACK)
        screen.blit(reac_label_surf, (x, y))
        x += 40
        y += reac_label_surf.get_height() * 1.5
        reac_txt_surf = debug_font.render(self.reaction_thoughts, True, BLACK)
        screen.blit(reac_txt_surf, (x, y))
    
    def draw_truthiness(self):
        x = self.x + 40
        y = self.y + 420
        truth_label_surf = debug_font_small.render('Is this for real?', True, BLACK)
        screen.blit(truth_label_surf, (x, y))
        x += 40
        y += truth_label_surf.get_height() * 1.5
        truth_str = str(self.truth)
        truth_txt_surf = debug_font.render(truth_str, True, BLACK)
        screen.blit(truth_txt_surf, (x, y))
        # were we forgiving or vindictive?
        x += truth_txt_surf.get_width() + 60
        if self.reaction.forgiving:
            c = pygame.Color(64, 200, 64)
            s = 'Forgiving'
        elif self.reaction.vindictive:
            c = pygame.Color(200, 64, 64)
            s = 'Vindictive'
        else:
            return
        perso_surf = debug_font_small.render(s, True, c)
        screen.blit(perso_surf, (x, y))
    
    def draw_speaker_affinity(self):
        # "consider the source"
        x = self.x + 40
        y = self.y + 370
        spk_label_surf = debug_font_small.render('Consider the source:', True, BLACK)
        screen.blit(spk_label_surf, (x, y))
        x += 40
        y += spk_label_surf.get_height() * 1.5
        spk_txt_surf = debug_font.render(self.speaker_thoughts, True, BLACK)
        screen.blit(spk_txt_surf, (x, y))
    
    def draw_final_affinity_changes(self):
        x = self.x + 40
        y = self.y + 470
        # student heads with +1, -2 etc
        result_label_surf = debug_font_small.render('Results:', True, BLACK)
        screen.blit(result_label_surf, (x, y))
        x += 160
        y += result_label_surf.get_height() * 2
        no_changes = True
        for student in self.reaction.aff_changes:
            change = self.reaction.aff_changes[student]
            # only show non-zero reactions
            if not change:
                continue
            no_changes = False
            student.face.draw_face_thumbnail((x, y), True, (0.5, 0.5))
            color = affinity_color(change)
            if change > 0:
                change = '+' + str(change)
            change = str(change)
            change_num_surf = debug_font.render(change, True, color)
            screen.blit(change_num_surf, (x + 40, y))
            y += 60
        # if no change, we are 'meh'
        if no_changes:
            nothin_surf = debug_font.render('Meh.', True, BLACK)
            screen.blit(nothin_surf, (x, y))
    
    def draw(self):
        self.draw_background()
        self.draw_titlebar()
        self.draw_gossip()
        # listener thoughts
        self.draw_thoughts_bg()
        # TODO: if speaker = listener, thoughts = "yep, i said that"; return
        # TODO: if subject = listener, thoughts = "this gossip is about me!"
        self.draw_truthiness()
        self.draw_gut_reaction()
        self.draw_speaker_affinity()
        # final result
        self.draw_final_affinity_changes()


class StudentFace(UIElement):
    "UI representation of a student"

    # if True, wait reaction_duration before continuing
    reaction_continue_delay = False
    # time in seconds the student will spend reacting
    # when a note reaches them
    reaction_duration = 2.0
    
    def __init__(self, ui, student, grid_x, grid_y):
        UIElement.__init__(self, ui)
        # save location in "desk grid" for note passing, etc
        self.desk_x = grid_x
        self.desk_y = grid_y
        self.student = student
        self.clique_color = None
        self.reacting_time = 0
        # each student has a unique, random mix of features
        self.face_style = feature_dir.random_style()
        # grid_x/y = place in NxN array of desks
        (x, y) = grid_to_screen(grid_x, grid_y)
        # use two shaded rectangles for a Zelda3-like top down perspective
        (w, h) = (DESK_WIDTH, DESK_DEPTH)
        self.desk_top = pygame.Rect(x, y, w, h)
        h = DESK_HEIGHT
        self.desk_bottom = pygame.Rect(x, y + DESK_DEPTH, w, h)
        # bounds = top and bottom together
        self.bounds = self.desk_top.union(self.desk_bottom)
        # increase bounds to include head
        face_rect = student_head_rect.copy()
        # but not 100% of head (image has lotsa blank space at borders)
        face_rect.topleft = (x, y - DESK_DISTANCE_Y * 0.8)
        self.bounds.union_ip(face_rect)
    
    def __str__(self):
        return "StudentFace: %s" % self.student.name
    
    def start_reacting(self):
        if log: print "%s started reacting" % self.student.name
        if self.reaction_continue_delay:
            self.reacting_time = time + (self.reaction_duration * 1000)
        # reaction will be visualized in self.draw_reaction()
    
    def stop_reacting(self):
        self.reacting_time = 0
        # student commits the affinity change for reals
        self.student.apply_reaction(self.student.reaction)
        if log: print "%s finished reacting" % self.student.name
        ui.finished_reacting()
    
    def update(self):
        # wait to see if reaction timer has elapsed ONLY IF it's possible
        if self.ui.state != UI_PASS_REACTING:
            return
        if not self.reaction_continue_delay:
            return
        if not self.reacting_time:
            return
        if self.reacting_time < time:
            self.stop_reacting()
    
    def is_adjacent(self, other):
        "returns True if we are within one desk of specified student (face)"
        # TODO: support other being a Student OR a StudentFace?
        x_range = range(self.desk_x - 1, self.desk_x + 2)
        y_range = range(self.desk_y - 1, self.desk_y + 2)
        if other.desk_x in x_range and other.desk_y in y_range:
            return True
        return False
    
    def click(self):
        # student can be dragged while creating gossip,
        # or picked as part of a route after that
        if self.ui.state == UI_GOSSIP:
            self.ui.drag_elements.append(self)
        elif self.ui.state == UI_NOTEROUTE:
            # add a Student to route, not a StudentFace
            self.ui.note.add_to_route(self.student)
    
    def unclick(self):
        # we might just have started hovering this frame,
        # and not registered a click()
        # route building is a multiple-clicks-quickly
        # operation, so account for that here
        if self.ui.state == UI_NOTEROUTE:
            self.ui.note.add_to_route(self.student)
    
    def hover(self):
        self.ui.selected_student = self.student
    
    def unhover(self):
        self.ui.update_selected_student()
    
    def draw_selection(self):
        (x, y) = self.desk_top.center
        highlight_circle = int(DESK_WIDTH * 0.6)
        pygame.draw.circle(screen, WHITE, (x, y), highlight_circle)
    
    def draw_desk(self):
        pygame.draw.rect(screen, MID_GREY, self.desk_bottom)
        pygame.draw.rect(screen, LIGHT_GREY, self.desk_top)    
    
    def draw_face_thumbnail(self, (x, y), centered=True, (scale_x, scale_y)=(0.5, 0.5), draw_name=False, name_color=BLACK):
        # TODO: StudentFace has become a monster!
        # consider a refactor in which this function is the core
        # of a class, and draw the other stuff in StudentFace
        # via other means
        "draws our face thumbnail at the specified coords"
        s = self.face_style.scaled_thumbnail(scale_x, scale_y)
        (w, h) = s.get_size()
        if centered:
            x -= w / 2
            y -= h / 2
        screen.blit(s, (int(x), int(y)))
        if not draw_name:
            return
        name_surf = debug_font_small.render(self.student.name, True, name_color)
        x += (w - name_surf.get_width()) / 2
        screen.blit(name_surf, (x, y + h))
    
    def draw_face(self, (x, y)):
        x, y = int(x), int(y)
        # affinity for selected student
        # in normal modes, just show affinity for selected student,
        # or blank expression if no selected student
        aff = 0
        if ui.selected_student and ui.state in [UI_VIEW, UI_GOSSIP, UI_NOTEROUTE]:
            # usually, show affinity for student
            aff = self.student.affinity[ui.selected_student]
            # if we're hovering a gossip box, show reaction to that student
            box = self.ui.hovering_class(gossip_box_classes)
            reac = self.student.reaction
            if box and reac:
                if self.ui.selected_student in reac:
                    aff = self.student.reaction.aff_changes[ui.selected_student]
                    # magnify effect for readability?
                    aff *= 2
                else:
                    # blank expression if we have no reaction to this
                    aff = 0
        # where to draw
        w = student_head_rect.width
        h = student_head_rect.height
        x -= w / 2
        y -= h
        forehead = vector.Vector2D(x, y)
        # get "look" direction from ui.focus
        dir = (self.ui.focus - forehead).normal()
        # look "straight ahead" when we're passing a note around
        if self.ui.state in [UI_PASSING, UI_PASS_REACTING]:
            dir = vector.ORIGIN
        # do same if we're the selected student
        if self.student is ui.selected_student:
            dir = vector.ORIGIN
        self.face_style.draw(screen, forehead, dir, aff)
    
    def draw_name(self, (x, y), inverted=False):
        "draws name on desk"
        if inverted:
            color = WHITE
        else:
            color = BLACK
        name = self.student.name
        if self.student is player:
            name += ' (You)'
        name_surf = debug_font.render(name, True, color)
        nx = x - name_surf.get_width() / 2
        ny = y - name_surf.get_height() / 2
        screen.blit(name_surf, (int(nx), int(ny)))
        # debug draw: total popularity
        if not draw_total_pop:
            return
        total = self.student.total_popularity()
        color = affinity_color(total)
        total_surf = debug_font.render(str(total), True, color)
        # draw total popularity off to side so heads don't cover it up
        x += DESK_WIDTH / 3
        y += total_surf.get_height() / 2
        screen.blit(total_surf, (int(x), int(y)))
        
    def draw_affinity(self):
        "draw this student's affinity for selected student"
        if ui.selected_student is None:
            return
        aff = self.student.affinity[ui.selected_student]
        # color text according to extremity of affinity
        color = affinity_color(aff)
        aff_surf = debug_font.render(str(aff), True, color)
        # change size of text according to same
        (w, h) = aff_surf.get_size()
        w += w * abs(aff) / 10.0
        h += h * abs(aff) / 10.0
        aff_surf = pygame.transform.smoothscale(aff_surf, (int(w), int(h)))
        (x, y) = self.desk_top.center
        x += student_head_rect.width / 2
        y -= student_head_rect.height / 2
        screen.blit(aff_surf, (x, y))
    
    def draw_reaction_preview(self):
        """
        draws this student's potential reaction (change in affinity)
        towards the current active gossip item
        """
        # student's last_reaction updates in Student.update()
        r = self.student.reaction
        # reaction will be "empty" if we have no feelings at all
        # about selected student!
        if len(r.aff_changes) == 0 or ui.selected_student not in r.aff_changes:
            return
        # keep an int and string version around
        reac = r.aff_changes[ui.selected_student]
        rstr = str(reac)
        if reac > 0:
            rstr = '+' + rstr
        # color according to affinity change
        color = affinity_color(reac)
        reac_surf = debug_font.render(rstr, True, color)
        (x, y) = self.desk_top.center
        x -= student_head_rect.width / 1.5
        y -= student_head_rect.height / 2
        screen.blit(reac_surf, (x, y))
        
    def draw_reaction(self):
        (x, y) = self.bounds.centerx, self.bounds.topright[1]
        screen.blit(thotbub_base_img, (x, y))
        y -= thotbub_img.get_height()
        off_y = thotbub_img.get_height() / 2
        # draw one or more bubble(s) for reaction
        no_reaction = True
        for r in self.student.reaction.aff_changes:
            # draw thought bubble even if we are 'meh'
            screen.blit(thotbub_img, (x, y))
            off_x = 16
            if self.student.reaction.aff_changes[r] == 0:
                continue
            no_reaction = False
            # format:
            # -1 (o_o) JERK!
            # ie affinity change + mini face + one/two word summary
            #
            # draw number            
            reac_num = self.student.reaction.aff_changes[r]
            reac_str = str(reac_num)
            if reac_num > 0:
                reac_str = "+%s" % reac_str
            num_surf = debug_font.render(reac_str, True, BLACK)
            screen.blit(num_surf, (x + off_x, y + off_y))
            # draw face to right of number
            off_x += num_surf.get_width() + 8
            face_thumb = r.face.face_style.scaled_thumbnail(0.25, 0.25)
            (face_w, face_h) = face_thumb.get_size()
            screen.blit(face_thumb, (x + off_x, y + off_y - face_h / 4))
            # draw summary to right of face
            off_x += face_w + 8
            if reac_num in thoughts.reaction_summary:
                s = thoughts.reaction_summary[reac_num]
            else:
                s = "%s" % r.name
            reac_surf = debug_font.render(s, True, BLACK)
            screen.blit(reac_surf, (x + off_x, y + off_y))
            # offset Y for any future bubbles
            y -= thotbub_img.get_height()
        # if we didn't have a reaction to anything in the note,
        # we are 'meh'
        if no_reaction:
            s = 'meh.'
            reac_surf = debug_font.render(s, True, BLACK)
            screen.blit(reac_surf, (x + off_x * 2, y + off_y))
    
    def refresh_clique_color(self):
        # color shows affinity within clique
        my_clique = self.student.clique
        # this is called from CliqueManager.update_cliques, so
        # our clique may have just disbanded
        if not my_clique or len(my_clique.students) < 2:
            self.clique_color = None
            if log: print 'no clique color for %s' % self.student
            return
        # color *= normalized rank within clique
        # (eg most popular student in clique = full color)
        rank = my_clique.rank(self.student)
        r = int(my_clique.color.r * rank)
        g = int(my_clique.color.g * rank)
        b = int(my_clique.color.b * rank)
        r = min(r, 255)
        g = min(g, 255)
        b = min(b, 255)
        # for some reason using color.r = to set doesn't work
        self.clique_color = pygame.Color(r, g, b, 255)
    
    def draw_clique_bg(self):
        # draw shaded rectangles to demarcate cliques
        r = self.bounds.inflate(DESK_DISTANCE_X, 16)
        pygame.draw.rect(screen, self.clique_color, r)
    
    def draw(self):
        if self.student.clique and self.clique_color:
            self.draw_clique_bg()
        # draw glow beneath desk only if selected
        if ui.selected_student is self.student:
            self.draw_selection()
        self.draw_desk()
        # if we're being dragged, draw centered on mouse cursor
        if self in ui.drag_elements:
            self.draw_face_thumbnail((mouse_x, mouse_y))
        # otherwise draw centered on desk
        else:
            self.draw_face(self.desk_top.center)
            self.draw_name(self.desk_bottom.center)
        # debug drawing of bounds
        #pygame.draw.rect(screen, (randint(0, 255), 0, 255), self.bounds)
        # player isn't an AI, their feelings aren't represented by a number :)
        if self.student is player:
            return
        if draw_affinity:
            self.draw_affinity()
        if self.student.reaction and ui.selected_student and \
            draw_gossip_result_preview:
            self.draw_reaction_preview()
        # visual to show "reacting"
        # (obsoleted by reaction dialogs)
        # if self.reaction and ui.state in [UI_PASS_REACTING, UI_PASSING]:
        #    self.draw_reaction()
        pass


class GossipNote():
    "visible representation of gossip that travels through the world"
    # pixels to move per frame
    interp_speed = 16
    # offset from top left corner of desk
    desk_offset = (DESK_WIDTH / 8, DESK_HEIGHT)
    route_color = BLUE
    trail_color = LIGHT_BLUE
    trail_width = 8
    
    # note is created when student hits "send" on gossip
    # route gets build, player clicks "pass" to set it going
    # note interps to each desk, student reacts, repeat until
    # final reaction is done, then delete
    def __init__(self, gossip):
        self.gossip = gossip
        # start at speaker's desk
        spk = gossip.speaker
        # route is list of students we will travel to, last to first
        self.route = [spk]
        # loc, dest and trail are screen coords
        (x, y) = grid_to_screen(spk.face.desk_x, spk.face.desk_y, self.desk_offset)
        self.loc = vector.Vector2D(x, y)
        # trail: list of screen points showing where we've traveled so far
        self.trail = []
        # dest: where we are currently interpolating to
        # (start at location while we're building route)
        self.dest = self.loc.copy()
    
    def set_next_dest_in_route(self):
        # remember next student to react now for when we arrive
        if log: print "GossipNote setting next destination in route..."
        next_student = self.route[0]
        (next_x, next_y) = (next_student.face.desk_x, next_student.face.desk_y)
        (next_x, next_y) = grid_to_screen(next_x, next_y, self.desk_offset)
        self.dest = vector.Vector2D(next_x, next_y)
        if log:
            if len(self.route) > 1:
                print "GossipNote next move: %s to %s" % (self.route[0].name, self.route[1].name)
            else:
                print "GossipNote at end of route"
    
    def update(self):
        # nothing here applies if we're not passing
        if ui.state != UI_PASSING:
            return
        # if we're at destination, pick next and tell UI to change state
        if self.loc == self.dest:
            if self.loc not in self.trail:
                self.trail.append(self.loc)
            ui.state_reacting()
            return
        # move towards destination at normal speed
        if vector.distance(self.loc, self.dest) >= self.interp_speed:
            dir = self.dest - self.loc
            move = vector.vectorMA(self.loc, dir.normal(), self.interp_speed)
            self.loc = move
        # snap to destination on final update
        else:
            if log: print "GossipNote arrived at %s" % self.route[0].name
            self.loc = self.dest.copy()
    
    def valid_for_route(self, student):
        "returns True if student is a  valid addition to current route"
        # can't already be in route
        if student in self.route:
            return False
        # can't be more than one desk away
        if not self.route[-1].face.is_adjacent(student.face):
            return False
        return True
    
    def add_to_route(self, student):
        "adds specified student to our route"
        if self.valid_for_route(student):
            self.route.append(student)
    
    def draw_route_options(self):
        # draw possible options if player is building a route
        for s in students:
            if self.valid_for_route(s):
                (x, y) = grid_to_screen(s.face.desk_x, s.face.desk_y, self.desk_offset)
                pygame.draw.circle(screen, MAGENTA, (x, y), 8, 1)
    
    def draw_route(self):
        plist = [self.loc.int_tuple()]
        for student in self.route:
            (x, y) = (student.face.desk_x, student.face.desk_y)
            p = grid_to_screen(x, y, self.desk_offset)
            plist.append(p)
        pygame.draw.lines(screen, self.route_color, False, plist, self.trail_width)
    
    def draw_trail(self):
        plist = []
        for point in self.trail:
            plist.append(point.int_tuple())
        plist.append(self.loc.int_tuple())
        if len(self.trail) < 1:
            return
        pygame.draw.lines(screen, self.trail_color, False, plist, self.trail_width)
    
    def draw(self):
        # show all points in route when building one
        if ui.state == UI_NOTEROUTE:
            self.draw_route_options()
            self.draw_route()
        # show our travel so far when passing around
        elif ui.state in [UI_PASSING, UI_PASS_REACTING]:
            self.draw_trail()
        # draw the note itself
        (w, h) = note_img.get_size()
        p = self.loc - (w / 2, h / 2)
        screen.blit(note_img, p.int_tuple())


class AffinityRankingBar(UIElement):
    
    "a 'totem pole' like vertical display of most to least popular students"
    
    face_size = 0.33
    # vertical space between faces
    face_spacing_y = 24
    # extra padding from right edge of screen
    x_margin = 8
    
    def __init__(self, ui):
        self.width = int(student_head_rect.height * self.face_size)
        self.x = screen.get_width() - self.width - self.x_margin
        self.y = 16
        UIElement.__init__(self, ui)
        self.rank_list = []
        self.color_list = []
    
    def update(self):
        if len(self.rank_list) == 0:
            self.refresh()
    
    def refresh(self):
        # list all students,
        # sort by aggregate affinity
        self.rank_list = students[:]
        self.rank_list.sort(cmp=lambda x,y: cmp(x.total_popularity(), y.total_popularity()))
        # most popular = first in list, most unpopular = last
        self.rank_list.reverse()
        # recompute list of colored boxes
        self.color_list = []
        for s in self.rank_list:
            if s.clique:
                self.color_list.append(s.face.clique_color)
            else:
                self.color_list.append(0)
    
    def draw(self):
        y = self.y
        # indicate that top = most popular
        top_surf = debug_font_small.render('Most Popular', True, LIGHT_GREY)
        screen.blit(top_surf, (self.x - top_surf.get_width() / 2, y))
        y += top_surf.get_height()
        i = 0
        # color all debug #s according to most liked or hated student
        most_pop = self.rank_list[0].total_popularity()
        least_pop = self.rank_list[-1].total_popularity()
        biggest_aff = max(most_pop, abs(least_pop))
        for s in self.rank_list:
            # draw background rect showing this student's clique, if any
            c = self.color_list[i]
            text_color = LIGHT_GREY
            if c:
                size = self.width + self.face_spacing_y
                box = pygame.Surface((size, size))
                box.fill(c)
                screen.blit(box, (self.x - self.x_margin, y))
                # use a different text color if we're on a colored background
                text_color = BLACK
            # draw face above that
            s.face.draw_face_thumbnail((self.x, y), False, (self.face_size, self.face_size), True, text_color)
            # make it clear where player is on totem pole
            if s is player:
                you_surf = debug_font.render('YOU', True, LIGHT_GREY)
                x = self.x - self.width * 1.5
                screen.blit(you_surf, (int(x), y + self.width / 2))
            # draw hard numbers if that debug mode is on
            if draw_total_pop:
                color = affinity_color(s.total_popularity(), biggest_aff)
                num_surf = debug_font.render(str(s.total_popularity()), True, color)
                screen.blit(num_surf, (self.x - self.width, y + size / 3))
            y += self.width + self.face_spacing_y
            i += 1
        # indicate that bottom = least popular
        bot_surf = debug_font_small.render('Least Popular', True, LIGHT_GREY)
        screen.blit(bot_surf, (self.x - bot_surf.get_width() / 2, y))


def load_assets():
    global note_img, thotbub_img, thotbub_base_img, thotbub_half_img
    global thotbub_huge_img
    global reaction_bubble_img, love_img, hate_img
    # misc assets
    if log: print "loading assets..."
    note_img = pygame.image.load('art/note.png')
    thotbub_img = pygame.image.load('art/thotbub.png')
    (w, h) = thotbub_img.get_size()
    thotbub_half_img = pygame.transform.smoothscale(thotbub_img, (int(w / 2), h))
    # size up thought bubble slightly until art gets redone
    thotbub_img = pygame.transform.smoothscale(thotbub_img, (int(w * 1.25), h))
    thotbub_base_img = pygame.image.load('art/thotbub_base.png')
    # assets for reaction dialog
    reaction_bubble_img = pygame.image.load('art/gossip_speech_thought.png')
    love_img = pygame.image.load('art/love_icon.png')
    hate_img = pygame.image.load('art/hate_icon.png')
    thotbub_huge_img = pygame.image.load('art/thotbub_huge.png')


def init():
    global feature_dir, student_head_rect
    load_assets()
    if log: print "building student face feature directory..."
    feature_dir = face.FaceFeatureDirectory()
    # it's useful to know the size of a student face, store it here
    student_head_rect = feature_dir.heads[0].stages[0].image.get_rect()
